"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.mergeWithMapping = exports.merge = void 0;
const xmldoc_1 = require("xmldoc");
const js_levenshtein_1 = __importDefault(require("js-levenshtein"));
const FUZZY_THRESHOLD = 0.2;
const STATE_INITIAL_XLF_2_0 = 'initial';
const STATE_INITIAL_XLF_1_2 = 'new';
function findCloseMatches(originUnit, destUnits) {
    const originText = toString(getSourceElement(originUnit));
    return destUnits
        .filter(n => n.type === 'element')
        .map(n => ({
        elem: n,
        score: (0, js_levenshtein_1.default)(originText, toString(getSourceElement(n))) / originText.length
    }))
        .filter(x => x.score < FUZZY_THRESHOLD)
        .sort((a, b) => a.score - b.score); // TODO test
}
function toString(...nodes) {
    return nodes.map(n => n.toString({ preserveWhitespace: true, compressed: true })).join('');
}
function collapseWhitespace(destSourceText) {
    return destSourceText.replace(/\s+/g, ' ');
}
function getUnits(doc, xliffVersion) {
    var _a, _b, _c;
    return xliffVersion === '2.0' ? (_a = doc.childNamed('file')) === null || _a === void 0 ? void 0 : _a.childrenNamed('unit') : (_c = (_b = doc.childNamed('file')) === null || _b === void 0 ? void 0 : _b.childNamed('body')) === null || _c === void 0 ? void 0 : _c.childrenNamed('trans-unit');
}
function getSourceElement(unit) {
    var _a, _b;
    // xliff 2.0: ./segment/source; xliff 1.2: ./source
    return (_b = (_a = unit.childNamed('segment')) === null || _a === void 0 ? void 0 : _a.childNamed('source')) !== null && _b !== void 0 ? _b : unit.childNamed('source');
}
function getTargetElement(unit) {
    var _a, _b;
    // xliff 2.0: ./segment/target; xliff 1.2: ./target
    return (_b = (_a = unit.childNamed('segment')) === null || _a === void 0 ? void 0 : _a.childNamed('target')) !== null && _b !== void 0 ? _b : unit.childNamed('target');
}
function createTargetElement(unit, xliffVersion) {
    // xliff 2.0: ./segment/target; xliff 1.2: ./target
    const parent = xliffVersion === '2.0' ? unit.childNamed('segment') : unit;
    const targetElement = new xmldoc_1.XmlDocument('<target></target>');
    parent.children.push(targetElement);
    updateFirstAndLastChild(parent);
    return targetElement;
}
function resetTranslationState(destUnit, xliffVersion, options) {
    var _a;
    if ((_a = options === null || options === void 0 ? void 0 : options.resetTranslationState) !== null && _a !== void 0 ? _a : true) {
        if (xliffVersion === '2.0') {
            destUnit.childNamed('segment').attr.state = (options === null || options === void 0 ? void 0 : options.sourceLanguage) ? 'final' : STATE_INITIAL_XLF_2_0;
        }
        else {
            const targetNode = destUnit.childNamed('target');
            if (targetNode) {
                targetNode.attr.state = (options === null || options === void 0 ? void 0 : options.sourceLanguage) ? 'final' : STATE_INITIAL_XLF_1_2;
            }
        }
    }
}
function isInitialState(destUnit, xliffVersion) {
    var _a, _b;
    if (xliffVersion === '2.0') {
        const state20 = (_a = destUnit.childNamed('segment')) === null || _a === void 0 ? void 0 : _a.attr.state;
        return state20 === undefined || state20 === STATE_INITIAL_XLF_2_0;
    }
    else {
        const state12 = (_b = destUnit.childNamed('target')) === null || _b === void 0 ? void 0 : _b.attr.state;
        return state12 === undefined || state12 === STATE_INITIAL_XLF_1_2;
    }
}
function revertApostrophes(s, revertApos) {
    return revertApos ? s.replace(/&apos;/g, '\'') : s;
}
function updateFirstAndLastChild(destUnit) {
    destUnit.firstChild = destUnit.children[0];
    destUnit.lastChild = destUnit.children[destUnit.children.length - 1];
}
function isUntranslated(destUnit, xliffVersion, destSourceText) {
    var _a;
    const targetElement = getTargetElement(destUnit);
    const destTargetText = toString(...(_a = targetElement === null || targetElement === void 0 ? void 0 : targetElement.children) !== null && _a !== void 0 ? _a : []);
    return isInitialState(destUnit, xliffVersion) && (!targetElement || destSourceText === destTargetText);
}
function createEmptyTarget(isXliffV2, srcLang, targetLang) {
    if (isXliffV2) {
        return `<?xml version="1.0" encoding="UTF-8"?>
<xliff version="2.0" xmlns="urn:oasis:names:tc:xliff:document:2.0" srcLang="${srcLang}" trgLang="${targetLang}">
  <file original="ng.template" id="ngi18n">
  </file>
</xliff>`;
    }
    else {
        return `<?xml version="1.0" encoding="UTF-8"?>
<xliff version="1.2" xmlns="urn:oasis:names:tc:xliff:document:1.2">
  <file source-language="${srcLang}" target-language="${targetLang}" datatype="plaintext" original="ng2.template">
    <body>
    </body>
  </file>
</xliff>`;
    }
}
function extractSourceLocale(translationSourceFile, isXliffV2) {
    var _a, _b;
    return (_b = (isXliffV2 ? translationSourceFile === null || translationSourceFile === void 0 ? void 0 : translationSourceFile.attr.srcLang : (_a = translationSourceFile === null || translationSourceFile === void 0 ? void 0 : translationSourceFile.childNamed('file')) === null || _a === void 0 ? void 0 : _a.attr['source-language'])) !== null && _b !== void 0 ? _b : 'en';
}
function extractTargetLocale(targetPath) {
    var _a, _b;
    return (_b = (_a = targetPath === null || targetPath === void 0 ? void 0 : targetPath.match(/\.([a-zA-Z-]+)\.xlf$/)) === null || _a === void 0 ? void 0 : _a[1]) !== null && _b !== void 0 ? _b : 'en';
}
function merge(inFilesContent, destFileContent, options, destFilePath) {
    const [mergedDestFileContent] = mergeWithMapping(inFilesContent, destFileContent, options, destFilePath);
    return mergedDestFileContent;
}
exports.merge = merge;
/**
 * syncs all elements from `source` to `dest` that are not in `ignoreElementNames`.
 * @param source
 * @param target
 * @param ignoreElementNames
 */
function syncOtherNodes(source, target, ...ignoreElementNames) {
    var _a;
    const targetNodesByName = new Map();
    target.children.filter((n) => n.type === 'element')
        .forEach(node => { var _a; return targetNodesByName.set(node.name, [...((_a = targetNodesByName.get(node.name)) !== null && _a !== void 0 ? _a : []), node]); });
    const sourceNodesByName = new Map();
    source.children.filter((n) => n.type === 'element')
        .forEach(node => { var _a; return sourceNodesByName.set(node.name, [...((_a = sourceNodesByName.get(node.name)) !== null && _a !== void 0 ? _a : []), node]); });
    // remove all nodes that are not in source:
    const removeSuperfluousTargetNodes = [...targetNodesByName.entries()].filter(([name]) => !ignoreElementNames.includes(name))
        .map(([name, nodes]) => { var _a, _b; return nodes.slice(0, nodes.length - ((_b = (_a = sourceNodesByName.get(name)) === null || _a === void 0 ? void 0 : _a.length) !== null && _b !== void 0 ? _b : 0)); })
        .flat();
    removeChildren(target, ...removeSuperfluousTargetNodes);
    let i = 0;
    let sourceOffset = 0;
    let targetOffset = 0;
    while (i + sourceOffset < source.children.length) {
        const targetElement = (_a = target.children) === null || _a === void 0 ? void 0 : _a[i + targetOffset];
        const sourceElement = source.children[i + sourceOffset];
        if (!(sourceElement.type === 'element') || ignoreElementNames.includes(sourceElement.name)) {
            sourceOffset++;
        }
        else if (i + targetOffset < target.children.length && (!((targetElement === null || targetElement === void 0 ? void 0 : targetElement.type) === 'element') || ignoreElementNames.includes(targetElement.name))) {
            targetOffset++;
        }
        else if ((targetElement === null || targetElement === void 0 ? void 0 : targetElement.type) === 'element' && targetElement.name === sourceElement.name) {
            targetElement.children = sourceElement.children;
            targetElement.attr = sourceElement.attr;
            updateFirstAndLastChild(targetElement);
            i++;
        }
        else {
            target.children.splice(i + targetOffset, 0, sourceElement);
            i++;
        }
        updateFirstAndLastChild(target);
    }
}
function getMinScoreId(bestMatchesIdToUnits) {
    let minScoreId;
    let minScore = Number.MAX_VALUE;
    bestMatchesIdToUnits.forEach((x, id) => {
        if (x.length) {
            const score = x[0].score;
            if (score < minScore) {
                minScore = score;
                minScoreId = id;
            }
        }
    });
    return minScoreId;
}
function mergeWithMapping(inFilesContent, destFileContent, options, destFilePath) {
    var _a, _b, _c, _d, _e, _f, _g;
    inFilesContent = Array.isArray(inFilesContent) ? inFilesContent : [inFilesContent];
    const inDocs = inFilesContent.map(inFileContent => new xmldoc_1.XmlDocument(inFileContent));
    const xliffVersion = (_a = inDocs[0].attr.version) !== null && _a !== void 0 ? _a : '1.2';
    const destDoc = new xmldoc_1.XmlDocument(destFileContent.match(/^[\n\r\s]*$/) ? createEmptyTarget(xliffVersion === '2.0', extractSourceLocale(inDocs[0], xliffVersion === '2.0'), extractTargetLocale(destFilePath)) : destFileContent);
    const excludeDocs = ((_b = options === null || options === void 0 ? void 0 : options.excludeFiles) !== null && _b !== void 0 ? _b : []).map(excludeFile => new xmldoc_1.XmlDocument(excludeFile));
    const destUnitsParent = xliffVersion === '2.0' ? destDoc.childNamed('file') : (_c = destDoc.childNamed('file')) === null || _c === void 0 ? void 0 : _c.childNamed('body');
    const excludeUnits = excludeDocs.map(excludeDoc => { var _a; return (_a = getUnits(excludeDoc, xliffVersion)) !== null && _a !== void 0 ? _a : []; }).flat(1);
    const excludeUnitsId = new Set(excludeUnits.map(unit => unit.attr.id));
    const inUnits = inDocs.map(inDoc => { var _a; return (_a = getUnits(inDoc, xliffVersion)) !== null && _a !== void 0 ? _a : []; }).flat(1).filter(inUnit => !excludeUnitsId.has(inUnit.attr.id));
    const inUnitsById = new Map(inUnits.map(unit => [unit.attr.id, unit]));
    const destUnitsById = new Map(((_d = getUnits(destDoc, xliffVersion)) !== null && _d !== void 0 ? _d : []).map(unit => [unit.attr.id, unit]));
    const allInUnitsWithoutDestinationUnit = inUnits.filter(u => !destUnitsById.has(u.attr.id));
    const allInUnitsWithDestinationUnit = inUnits.filter(u => destUnitsById.has(u.attr.id));
    // collect (potentially) obsolete units (defer actual removal to allow for fuzzy matching..):
    const removeNodes = getUnits(destDoc, xliffVersion).filter(destUnit => !inUnitsById.has(destUnit.attr.id));
    const idMapping = {};
    /** Syncs `unit` to `destUnit` or adds `unit` as new, if `destUnit` is not given. */
    function handle(unit, destUnit) {
        var _a, _b, _c;
        const unitSource = getSourceElement(unit);
        const unitSourceText = toString(...unitSource.children);
        if (destUnit) {
            const destSource = getSourceElement(destUnit);
            const destSourceText = toString(...destSource.children);
            const originTarget = getTargetElement(unit);
            const destTarget = getTargetElement(destUnit);
            if (((_a = options === null || options === void 0 ? void 0 : options.collapseWhitespace) !== null && _a !== void 0 ? _a : true) ? collapseWhitespace(destSourceText) !== collapseWhitespace(unitSourceText) : destSourceText !== unitSourceText) {
                destSource.children = unitSource.children;
                if ((options === null || options === void 0 ? void 0 : options.sourceLanguage) || ((options === null || options === void 0 ? void 0 : options.syncTargetsWithInitialState) === true && isUntranslated(destUnit, xliffVersion, destSourceText))) {
                    const targetElement = destTarget !== null && destTarget !== void 0 ? destTarget : createTargetElement(destUnit, xliffVersion);
                    targetElement.children = unitSource.children;
                }
                updateFirstAndLastChild(destSource);
                resetTranslationState(destUnit, xliffVersion, options);
                console.debug(`update element with id "${unit.attr.id}" with new source: ${toString(...destSource.children)} (was: ${destSourceText})`);
            }
            else if (originTarget && !destTarget) {
                const sourceIndex = destUnit.children.indexOf(destSource);
                destUnit.children.splice(sourceIndex + 1, 0, originTarget);
            }
            if (destUnit.attr.id !== unit.attr.id) {
                console.debug(`matched unit with previous id "${destUnit.attr.id}" to new id: "${unit.attr.id}"`);
                idMapping[destUnit.attr.id] = unit.attr.id;
                removeNodes.splice(removeNodes.indexOf(destUnit), 1);
                destUnit.attr.id = unit.attr.id;
                resetTranslationState(destUnit, xliffVersion, options);
            }
            syncOtherNodes(unit, destUnit, 'source', 'target', 'segment');
            updateFirstAndLastChild(destUnit);
        }
        else {
            console.debug(`adding element with id "${unit.attr.id}"`);
            if (!getTargetElement(unit)) {
                if ((options === null || options === void 0 ? void 0 : options.newTranslationTargetsBlank) !== 'omit') {
                    const shouldBeBlank = ((_b = options === null || options === void 0 ? void 0 : options.newTranslationTargetsBlank) !== null && _b !== void 0 ? _b : false) && !((_c = options === null || options === void 0 ? void 0 : options.sourceLanguage) !== null && _c !== void 0 ? _c : false);
                    const targetNode = new xmldoc_1.XmlDocument(`<target>${shouldBeBlank ? '' : unitSourceText}</target>`);
                    if (xliffVersion === '2.0') {
                        const segmentSource = unit.childNamed('segment');
                        segmentSource.children.push(targetNode);
                    }
                    else {
                        const sourceIndex = unit.children.indexOf(unitSource);
                        unit.children.splice(sourceIndex + 1, 0, targetNode);
                    }
                }
                resetTranslationState(unit, xliffVersion, options);
            }
            destUnitsParent.children.push(unit);
            updateFirstAndLastChild(destUnitsParent);
        }
    }
    for (const unit of allInUnitsWithDestinationUnit) {
        handle(unit, destUnitsById.get(unit.attr.id));
    }
    if ((_e = options === null || options === void 0 ? void 0 : options.fuzzyMatch) !== null && _e !== void 0 ? _e : true) {
        const bestMatchesIdToUnits = new Map(allInUnitsWithoutDestinationUnit.map((inUnit) => [inUnit.attr.id, findCloseMatches(inUnit, removeNodes)]));
        while (bestMatchesIdToUnits.size) {
            const inUnitId = (_f = getMinScoreId(bestMatchesIdToUnits)) !== null && _f !== void 0 ? _f : [...bestMatchesIdToUnits.keys()][0];
            const bestMatch = (_g = bestMatchesIdToUnits.get(inUnitId)[0]) === null || _g === void 0 ? void 0 : _g.elem;
            handle(inUnitsById.get(inUnitId), bestMatch);
            bestMatchesIdToUnits.delete(inUnitId);
            if (bestMatch) {
                bestMatchesIdToUnits.forEach(x => {
                    const i = x.findIndex(y => y.elem === bestMatch);
                    if (i >= 0) {
                        x.splice(i, 1);
                    }
                });
            }
        }
    }
    else {
        for (const unit of allInUnitsWithoutDestinationUnit) {
            handle(unit, undefined);
        }
    }
    console.debug(`removing ${removeNodes.length} ids: ${removeNodes.map(n => n.attr.id).join(', ')}`);
    removeChildren(destUnitsParent, ...removeNodes);
    // retain xml declaration:
    const xmlDecMatch = destFileContent.match(/^<\?xml [^>]*>\s*/i);
    const xmlDeclaration = xmlDecMatch ? xmlDecMatch[0] : '';
    const mergedContent = xmlDeclaration + revertApostrophes(destDoc.toString({
        preserveWhitespace: true,
        compressed: true
    }), !(options === null || options === void 0 ? void 0 : options.replaceApostrophe));
    return [mergedContent, idMapping];
}
exports.mergeWithMapping = mergeWithMapping;
/**
 * Automatically removes whitespace text nodes before children.
 *
 * @param node
 * @param children
 */
function removeChildren(node, ...children) {
    const removeIndexes = new Set(node.children.map((c, i) => children.indexOf(c) >= 0 ? i : null).filter(x => x !== null));
    node.children = node.children.filter((c, i) => !removeIndexes.has(i) && (!removeIndexes.has(i + 1) || !isWhiteSpace(c)));
    updateFirstAndLastChild(node);
}
function isWhiteSpace(node) {
    return node.type === 'text' && !!node.text.match(/^\s*$/);
}
